/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
01/11/01 - Bobby G. Vinyard (Message)
  - Added four additional modifiers:  alt     alt key
                                      lwin    left win button
                                      rwin    right win button
                                      apps    button next to the right win button
01/10/01 - Bobby G. Vinyard (Message)
  - Removed DesktopLeftClick, DesktopRightClick, DesktopMiddleClick,
    DesktopX1Click, DesktopX2Click, !DesktopSetRightClick, !DesktopSetLeftClick,
    !DesktopSetMiddleClick, !DesktopSetX1Click, !DesktopSetX2Click
  - Fixed desktop getting focus when using xmouse
  - Addednew support for key modifiers used as the following:
    *desktop [_tmain button]+[modifier]+[modifier]+.. command
    Modifiers are:  lbutton     left mouse button up
                    rbutton     right mouse button up
                    mbutton     middle mouse button up
                    x1button    x1 mouse button up
                    x2button    x2 mouse button up
                    ctrl        control key down
                    shift       shift key down
                    
    Examples:
    *desktop rbutton !popup // Right button up displays popup
    *desktop lbutton !about // Left button up displays about box
    *desktop lbutton+ctrl !recycle // Left button up while holding down control key
                                      recycles LiteStep
    *desktop rbutton+ctrl+shift !quit // You can figure it out =]
    
    Note: you can do *desktop rbutton+lbutton !about and it will display the
    about box when you hold down the lbutton and click the rbutton, *but* releasing 
    the lbutton will fire off the command set by *desktop lbutton
  - Added !DesktopSetClick, you use this in the same way as the step.rc config, just
    minus the "*desktop"
    !DesktopSetClick lbutton+ctrl !recycle 
12/08/00 - Bobby G. Vinyard (Message)
  - Added XButton support, DesktopX1Click, DesktopX2Click, !SetDesktopX1Click,
    !SetDesktopX2Click
12/07/00 - Bobby G. Vinyard (Message)
  - Cleaned up the code some, added LM_REFRESH handling
12/06/00 - Bobby G. Vinyard (Message)
  - Added SetDesktopAreaMainMonitorOnly, SetDesktopArea<int>, SDALeft<int>
    SDATop<int>, SDABottom<int>, SDARight<int> see desktop2.tx t for details.
11/28/00 - Bobby G. Vinyard (Message)
  - Added multimonitor support
11/23/00 - Joachim Calvert (NeXTer)
  - The !DesktopSetArea command is finally working, I must have been very
    tired when I wrote that code. Not only did it call the wrong function,
    but also registered the wrong command. That noone hasn't noticed before
    now is pretty remarkable
11/08/00 - Joachim Calvert (NeXTer)
  - Oups... Setting the parent to the desktop made it show up in the task
    list. Back to the drawing board
11/07/00 - Joachim Calvert (NeXTer)
  - Changed the parent window to the desktop and removed the clipping.
    This should (theoretically) make it easier to adapt LS to a multi-monitor
    setup
11/01/00 - Joachim Calvert (NeXTer)
  - Added !DesktopSetArea left top right bottom, the rules are the same as
    for the normal settings
05/21/00 - Joachim Calvert (NeXTer)
  - Removed some superflous code, and fixed some other minor issues.
04/20/00 - Joachim Calvert (NeXTer)
  - Some minor fixes
04/19/00 - Joachim Calvert (NeXTer)
  - Now uses Window (in lswinbase) as base class, should simplify the process
    of writing C++ based modules somewhat
04/17/00 - Joachim Calvert (NeXTer)
  Made the code more or less completely streamlined. The size has suffered
    slightly, but the readability is much improved.
04/16/00 - Bobby G. Vinyard (Message)
  Changed the desktop clicks to use LSExecute instead of ParseBangCommand
    This results in desktop2 not needing to parse the string before calling
    LSExecute, removed some unneeded variables and functions
    Can now do something like DesktopLeftClick cmd.exe to cause the left click
    to open a command window in WinNT
04/14/00 - Bobby G. Vinyard (Message)
  Made right click default to !popup
04/14/00 - Joachim Calvert (NeXTer)
  The desktop area is now dynamically set on resolution changes.
04/13/00 - Bobby G. Vinyard (Message)
  Reimplemented desktop clicks to use a function for setting their variables
    (Thanks MrJukes and Dem0sh for your help)
  Added three new bang commands:
    !SetDesktopLeftClick
    !SetDesktopRightClick
    !SetDesktopMiddleClick
    These allow you to change what the mouse clicks do on the fly
    e.g. !SetDesktopLeftClick !popup x=0 y=0
04/10/00 Bobby G. Vinyard (Message)
  Remove autorun functionality (was our understanding that this was flakey
    and that a 3rd party module is available that better provides this
    functionality, may be reimplemented in the future, but I see no reason
    for it to be a desktop function)
  Did some more house cleaning (i.e. removed unused variables, added proper
    headers, proper format)
04/06/00 Bobby G. Vinyard (Message) & noodge
  Went ahead and added a basic implementation of DesktopLeftClick <command>,
  DesktopMiddleClick <command>, DesktopRightClick <command> (Might look into
    streamling the functions soon when I finish streamling desktop2)
04/01/00 Bobby G. Vinyard (Message)
  Removed some unused variable (ok, well really just commented them out until
    this build is throughly tested)
  Did some more work on setting the desktop area, hopefully this is fixed now
****************************************************************************/

#include "Desktop.h"

const char szAppName[] = "DesktopBackgroundClass"; // Our window class, etc

const char rcsRevision[] = "0.3 $"; // Our Version
const char rcsId[] = "IconDesktop, v 0.3 2003/04/24 15:43:24 Sci Exp $"; // The Full RCS ID.

Desktop *desktop; // The module
int numMonitors;
bool multipleSDA;

BOOL CALLBACK MultipleDesktopAreaEnumProc(HMONITOR hMonitor, HDC gdc, LPRECT prc, LPARAM lParam);
BOOL CALLBACK DesktopAreaEnumProc(HMONITOR hMonitor, HDC gdc, LPRECT prc, LPARAM lParam);
BOOL CALLBACK DesktopAreaResetEnumProc(HMONITOR hMonitor, HDC gdc, LPRECT prc, LPARAM lParam);

#ifdef _DEBUG
char szdebugBuffer[4096];
#endif // _DEBUG

//=========================================================
// Initialization and cleanup
//=========================================================

int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	int code;

	Window::init(dllInst);

	desktop = new Desktop(ParentWnd, code);

	return code;
}


void quitModule(HINSTANCE dllInst)
{
	delete desktop;
}


//=========================================================
// Bang commands
//=========================================================

void BangSetClick(HWND caller, LPCSTR args)
{
	desktop->bangSetClick(args);
}


void BangDesktopSetArea(HWND caller, LPCSTR args)
{
	desktop->bangSetArea(args);
}

static int ParseIconLocation(LPSTR szIconFile)
{
	LPSTR pszComma = strrchr(szIconFile, ',');

	if (!pszComma)
		return -1;

	*pszComma = 0;
	return atoi(pszComma + 1);
}


HICON GetIconFromPathEx(LPCSTR pszPath, int cxDesired, int cyDesired)
{
	char szIconFile[MAX_PATH];
	int nIconIndex = -1;
	LPCSTR pszExt;

    if (strlen(pszPath) > 0 && pszPath[0])
    {
        if (stricmp(pszPath, ".none") == 0)
        {
            return NULL;
        }
        
        strncpy(szIconFile, pszPath, MAX_PATH);
        nIconIndex = ParseIconLocation(szIconFile);
        
        pszExt = strrchr(szIconFile, '.');
            
        if (pszExt && stricmp(pszExt, ".ico") == 0)
        {
            return (HICON) LoadImage(NULL, szIconFile, IMAGE_ICON,
                cxDesired, cyDesired, LR_LOADFROMFILE);
        }
        else if (nIconIndex >= 0)
        {
            HICON hIcon;
                
            ExtractIconEx(szIconFile, nIconIndex,
                (cxDesired <= 24 ? NULL : &hIcon),
                (cxDesired <= 24 ? &hIcon : NULL), 1);
                
            return hIcon;
        }
        else
        {
            SHFILEINFO shfi;
            char pszFileName[MAX_PATH];
            LPCSTR pszFile = szIconFile;
            UINT uFlags = NULL;
                
            if (strchr(szIconFile, '*'))
            {
                PathStripPath(szIconFile);
                uFlags |= SHGFI_USEFILEATTRIBUTES;
            }
            else if (!PathFileExists(szIconFile))
            {
                FindExecutable(szIconFile, NULL, pszFileName);
                pszFile = pszFileName;
            }
                
            SHGetFileInfo(pszFile, 0, &shfi, sizeof(SHFILEINFO),
                SHGFI_ICON | (cxDesired <= 24 ? SHGFI_SMALLICON : 0) | uFlags);
          
            return shfi.hIcon;
        }
    }
    else
    {
        // get generic folder icon
        SHFILEINFO shfi;
        
        SHGetFileInfo("*", FILE_ATTRIBUTE_DIRECTORY, &shfi, sizeof(SHFILEINFO),
            SHGFI_USEFILEATTRIBUTES | SHGFI_ICON | (cxDesired <= 24 ? SHGFI_SMALLICON : 0));
        
        return shfi.hIcon;
    }
}

//=========================================================
// Module code
//=========================================================

//---------------------------------------------------------
// Module constructor
//---------------------------------------------------------
Desktop::Desktop(HWND parentWnd, int& code):
		Window(szAppName)
{
	int msgs[] = {LM_GETREVID, LM_REPAINT, LM_REFRESH, 0};

	screenLeft = SCREEN_LEFT;
	screenTop = SCREEN_TOP;
	screenWidth = SCREEN_WIDTH;
	screenHeight = SCREEN_HEIGHT;

	if(!createWindow(WS_EX_TOOLWINDOW, szAppName, WS_POPUP, screenLeft, screenTop, screenWidth, screenHeight, parentWnd)) {
		MessageBox(hWnd, "Unable to create window", szAppName, MB_SETFOREGROUND);
		code = 1;
		return;
	}

	SetWindowPos(hWnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE | SWP_SHOWWINDOW | SWP_NOSENDCHANGING);

	SendMessage(hParent, LM_REGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);
	SetWindowLong(hWnd, GWL_USERDATA, magicDWord);

	SetClassLong(hWnd, GCL_STYLE, GetClassLong(hWnd, GCL_STYLE) | CS_DBLCLKS);

	AddBangCommand("!DesktopSetClick", BangSetClick);
	AddBangCommand("!DesktopSetArea", BangDesktopSetArea);

	setupDesktop();

	code = 0;
}


Desktop::~Desktop()
{
    int msgs[] = {LM_GETREVID, LM_REPAINT, LM_REFRESH, 0};

	SendMessage(hParent, LM_UNREGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);

	RemoveBangCommand("!DesktopSetClick");
	RemoveBangCommand("!DesktopSetArea");

	resetMinMax();

	destroyWindow();

	cleanDesktop();
}


// --------------------------------------------------------
// Sets maximized windows size
// --------------------------------------------------------
void Desktop::setMinMax(void)
{
	RECT r = {screenLeft, screenTop, screenWidth, screenHeight};
	multipleSDA = false;

	EnumDisplayMonitors(NULL, NULL, MultipleDesktopAreaEnumProc, 0);
	if (multipleSDA)
	{
		if (setDesktopArea)
		{
			RECT rcMain = {0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN)};
			if (sdaRight <= 0)
				r.right = rcMain.right + sdaRight;
			else
				r.right = sdaRight;
			if (sdaBottom <= 0)
				r.bottom = rcMain.bottom + sdaBottom;
			else
				r.bottom = sdaBottom;
			SetDesktopArea(r.left, r.top, r.right, r.bottom);
		}
		return ;
	}

	if (setDesktopArea)
	{
		r.left = sdaLeft;
		r.top = sdaTop;
		if (sdaRight <= 0)
			r.right = r.right + sdaRight;
		else
			r.right = sdaRight;
		if (sdaBottom <= 0)
			r.bottom = r.bottom + sdaBottom;
		else
			r.bottom = sdaBottom;
	}
	if (setDesktopAreaMainMonitorOnly)
	{
		RECT rcMain = {0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN)};
		if (sdaRight <= 0)
			r.right = rcMain.right + sdaRight;
		else
			r.right = sdaRight;
		if (sdaBottom <= 0)
			r.bottom = rcMain.bottom + sdaBottom;
		else
			r.bottom = sdaBottom;
		EnumDisplayMonitors(NULL, NULL, DesktopAreaEnumProc, (LPARAM)&rcMain);
	}

#ifdef _DEBUG
	sprintf(szdebugBuffer, "left: %i top: %i right: %i bottom: %i",
	          r.left, r.top, r.right, r.bottom);
	_LSDEBUG2("New Work Area", szdebugBuffer);
#endif // _DEBUG

	SetDesktopArea(r.left, r.top, r.right, r.bottom);
}


//---------------------------------------------------------
// Reset to old maximized windows size
//---------------------------------------------------------
void Desktop::resetMinMax(void)
{
	EnumDisplayMonitors(NULL, NULL, DesktopAreaResetEnumProc, 0);
}


void Desktop::setupDesktop()
{
	FILE* f;
	char szDisplayChange[MAX_LINE_LENGTH+1];

	numMonitors = 1;

	setDesktopArea = GetRCBool("SetDesktopArea", true) != FALSE;
	setDesktopAreaMainMonitorOnly = GetRCBool("SetDesktopAreaMainMonitorOnly", true) != FALSE;
	sdaLeft = GetRCInt("SDALeft", 0);
	sdaRight = GetRCInt("SDARight", 0);
	sdaTop = GetRCInt("SDATop", 0);
	sdaBottom = GetRCInt("SDABottom", 0);
	szDisplayChange[0] = '\0';
	GetRCString("DesktopOnDisplayChange", szDisplayChange, "", MAX_LINE_LENGTH);
	bangDisplayChange = szDisplayChange;

	// Set the default right click to !popup
	bangClicks.insert(mapClicks::value_type((UINT)MOD_RBUTTON, "!popup"));

	f = LCOpen(NULL);
	if (f)
	{
		char	buffer[4096];
		char	token1[4096], token2[4096], extra_text[4096];
		char*	tokens[2];

		tokens[0] = token1;
		tokens[1] = token2;

		buffer[0] = 0;

		while (LCReadNextConfig (f, "*Desktop", buffer, sizeof (buffer)))
		{
			int count;

			token1[0] = token2[0] = extra_text[0] = '\0';

			count = LCTokenize (buffer, tokens, 2, extra_text);

			if (count == 2)
			{

				UINT nModifier = 0;
				nModifier = parseLine(token2);

				mapClicks::iterator itClicks;
				itClicks = bangClicks.find(nModifier);
				if (itClicks != bangClicks.end())
				{
					bangClicks.erase(itClicks);
				}
				bangClicks.insert(mapClicks::value_type(nModifier, extra_text));
			}
		}
	}
	LCClose(f);

	// Icon settings
	_iconList = NULL;
	_iconsLoaded = FALSE;
	_iconX = GetRCInt("DesktopIconX", 0);
	_iconY = GetRCInt("DesktopIconY", 0);
	_iconSize = GetRCInt("DesktopIconSize", 32);
	_iconSpacingX = GetRCInt("DesktopIconSpacingX", 36);
	_iconSpacingY = GetRCInt("DesktopIconSpacingY", 42);

	GetRCString("DesktopIconFolder", _iconFolder, "C:\\", MAX_PATH);
	if(_iconFolder[strlen(_iconFolder)-1] != '\\') {
		strcat(_iconFolder, "\\");
	}
	_iconHideExtension = GetRCBoolDef("DesktopIconHideExtension", TRUE);

	char buffer[MAX_LINE_LENGTH];

	GetRCString("DesktopIconTextBackground", buffer, "", MAX_LINE_LENGTH);
	if(strlen(buffer)) {
		_iconTextBackground = CreateSolidBrush(GetRCColor("DesktopIconTextBackground", RGB(255,255,255)));
	}
	else {
		_iconTextBackground = NULL;
	}
	_iconTextSpacing = GetRCInt("DesktopIconTextSpacing", 5);
	_iconTextHeight = GetRCInt("DesktopIconTextHeight", 28);
	_iconTextWidth = GetRCInt("DesktopIconTextWidth", 64);

	_iconTextColor = GetRCColor("DesktopIconTextFontColor", RGB(0,0,0));

	GetRCString("DesktopIconTextFont", buffer, "Arial", MAX_LINE_LENGTH);
	int fontSize = GetRCInt("DesktopIconTextFontSize", 14);
	BOOL fontItalic = GetRCBoolDef("DesktopIconTextFontItalic", FALSE);
	BOOL fontBold = GetRCBoolDef("DesktopIconTextFontBold", FALSE);
	if(!(_iconTextFont = CreateFont(fontSize, 0, 0, 0, fontBold ? FW_BOLD : FW_NORMAL, fontItalic, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, buffer))) {
		LSLogPrintf(LOG_ERROR, "IconDesktop", "Could not load font \"%s\"", buffer);
	}

	GetRCString("DesktopIconDirection", buffer, "", MAX_LINE_LENGTH);
	if(!stricmp(buffer, "down")) {
		_iconDirection = DD_DOWN;
	}
	else if(!stricmp(buffer, "left")) {
		_iconDirection = DD_LEFT;
	}
	else if(!stricmp(buffer, "up")) {
		_iconDirection = DD_UP;
	}
	else {
		_iconDirection = DD_RIGHT;
	}

	GetRCString("DesktopIconWrapDirection", buffer, "", MAX_LINE_LENGTH);
	if(_iconDirection == DD_DOWN || _iconDirection == DD_UP) {
		if(!stricmp(buffer, "left")) {
			_iconWrapDirection = DD_LEFT;
		}
		else {
			_iconWrapDirection = DD_RIGHT;
		}
	}
	else {
		if(!stricmp(buffer, "up")) {
			_iconWrapDirection = DD_UP;
		}
		else {
			_iconWrapDirection = DD_DOWN;
		}
	}
	_iconWrapCount = max(GetRCInt("DesktopIconWrapCount", 10), 1);

	_iconChangeHandle = FindFirstChangeNotification(_iconFolder, FALSE, FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME);
	if(_iconChangeHandle == INVALID_HANDLE_VALUE) {
		LSLogPrintf(LOG_ERROR, "IconDesktop", "Could not create change notification handle");
	}
	SetTimer(hWnd, 0, GetRCInt("DesktopIconUpdateInterval", 500), NULL);

	setMinMax();
}

UINT Desktop::parseLine(LPSTR line)
{
	char* tmp;
	UINT nModifier = 0;

	tmp = strtok(line, "+");

	if (!strcmpi(tmp, "lbutton"))
		nModifier |= MOD_LBUTTON;
	if (!strcmpi(tmp, "mbutton"))
		nModifier |= MOD_MBUTTON;
	if (!strcmpi(tmp, "rbutton"))
		nModifier |= MOD_RBUTTON;
	if (!strcmpi(tmp, "x1button"))
		nModifier |= MOD_XBUTTON1;
	if (!strcmpi(tmp, "x2button"))
		nModifier |= MOD_XBUTTON2;
    
	if (nModifier == 0)
		return 0;

	tmp = strtok(NULL, "+");

	while (tmp)
	{
		if (!strcmpi(tmp, "lbutton"))
			nModifier |= MK_LBUTTON;
		if (!strcmpi(tmp, "mbutton"))
			nModifier |= MK_MBUTTON;
		if (!strcmpi(tmp, "rbutton"))
			nModifier |= MK_RBUTTON;
		if (!strcmpi(tmp, "x1button"))
			nModifier |= MK_XBUTTON1;
		if (!strcmpi(tmp, "x2button"))
			nModifier |= MK_XBUTTON2;
		if (!strcmpi(tmp, "ctrl"))
			nModifier |= MK_CONTROL;
		if (!strcmpi(tmp, "shift"))
			nModifier |= MK_SHIFT;
		if (!strcmpi(tmp, "alt"))
			nModifier |= MK_MENU;
		if (!strcmpi(tmp, "apps"))
			nModifier |= MK_APPS;
		if (!strcmpi(tmp, "lwin"))
			nModifier |= MK_LWIN;
		if (!strcmpi(tmp, "rwin"))
			nModifier |= MK_RWIN;
		tmp = strtok(NULL, "+");
	}

	return nModifier;
}
void Desktop::cleanDesktop()
{
	bangClicks.clear();

	_deleteIcons();

	if(_iconTextFont) {
		DeleteObject(_iconTextFont);
	}
	if(_iconTextBackground) {
		DeleteObject(_iconTextBackground);
	}

	KillTimer(hWnd, 0);
	FindCloseChangeNotification(_iconChangeHandle);
}
//=========================================================
// Registered messages
//=========================================================

void Desktop::windowProc(Message& message)
{
	BEGIN_MESSAGEPROC
	//MESSAGE(onCreate,            WM_CREATE)
	//MESSAGE(onDestroy,           WM_DESTROY)
	MESSAGE(onActivate, WM_ACTIVATE)
	MESSAGE(onDisplayChange, WM_DISPLAYCHANGE)
	MESSAGE(onEndSession, WM_ENDSESSION)
	MESSAGE(onEndSession, WM_QUERYENDSESSION)
	MESSAGE(onEraseBkgnd, WM_ERASEBKGND)
	MESSAGE(onGetRevId, LM_GETREVID)
	MESSAGE(onKeyMessage, WM_KEYDOWN)
	MESSAGE(onKeyMessage, WM_KEYUP)
	MESSAGE(onKeyMessage, WM_HOTKEY)
	MESSAGE(onMouseActivate, WM_MOUSEACTIVATE)
	MESSAGE(onMouseButtonDown, WM_LBUTTONDOWN)
	MESSAGE(onMouseButtonDown, WM_MBUTTONDOWN)
	MESSAGE(onMouseButtonDown, WM_RBUTTONDOWN)
	MESSAGE(onMouseButtonDown, WM_XBUTTONDOWN)
	MESSAGE(onMouseButtonUp, WM_LBUTTONUP)
	MESSAGE(onMouseButtonUp, WM_MBUTTONUP)
	MESSAGE(onMouseButtonUp, WM_RBUTTONUP)
	MESSAGE(onMouseButtonUp, WM_XBUTTONUP)
	MESSAGE(onMouseButtonDbl, WM_LBUTTONDBLCLK)
	MESSAGE(onMouseButtonDbl, WM_MBUTTONDBLCLK)
	MESSAGE(onMouseButtonDbl, WM_RBUTTONDBLCLK)
	MESSAGE(onMouseButtonDbl, WM_XBUTTONDBLCLK)
	MESSAGE(onSysCommand, WM_SYSCOMMAND)
	MESSAGE(onWindowPosChanging, WM_WINDOWPOSCHANGING)
	MESSAGE(onRefresh, LM_REFRESH)
	MESSAGE(onPaint, WM_PAINT)
	MESSAGE(onTimer, WM_TIMER)
	END_MESSAGEPROC
}

void Desktop::_deleteIcons()
{
	DESKTOPICON *loop = _iconList;
	while(loop) {
		loop = _iconList->next;
		if(_iconList->icon) {
			DeleteObject(_iconList->icon);
		}
		free(_iconList);
		_iconList = loop;
	}
	_iconsLoaded = FALSE;
}

void Desktop::_loadIcons()
{
	WIN32_FIND_DATA fd;
	HANDLE search;
	char filter[MAX_PATH] = "\0";
	BOOL finished = FALSE;
	int i = 0, x = 0, y = 0;

	strcpy(filter, _iconFolder);
	strcat(filter, "*.*");

	search = FindFirstFile(filter, &fd);
	if(search == INVALID_HANDLE_VALUE) {
		LSLogPrintf(LOG_DEBUG, "IconDesktop", "No files found.");
	}

	while(!finished) {
		if(strcmp(fd.cFileName, ".") && strcmp(fd.cFileName, "..")) {
			DESKTOPICON *icon = (DESKTOPICON *)malloc(sizeof(DESKTOPICON));
			memset(icon, 0, sizeof(DESKTOPICON));
			if(_iconList) {
				DESKTOPICON *last = _iconList;
				while(last->next) {
					last = last->next;
				}
				last->next = icon;
			}
			else {
				_iconList = icon;
			}

			strcpy(icon->filename, _iconFolder);
			strcat(icon->filename, fd.cFileName);
			strcpy(icon->title, fd.cFileName);
			if(_iconHideExtension) {
				PathRemoveExtension(icon->title);
			}

			switch(_iconDirection) {
				case DD_RIGHT:
				case DD_LEFT:
					x = (i % _iconWrapCount);
					y = (i / _iconWrapCount);
					if(_iconWrapDirection == DD_UP) {
						y = -y;
					}
					if(_iconDirection == DD_LEFT) {
						x = -x;
					}
					break;

				case DD_DOWN:
				case DD_UP:
					x = (i / _iconWrapCount);
					y = (i % _iconWrapCount);
					if(_iconWrapDirection == DD_LEFT) {
						x = -x;
					}
					if(_iconDirection == DD_UP) {
						y = -y;
					}
					break;

			}

			icon->x = _iconX + ((_iconSize + _iconSpacingX) * x);
			icon->y = _iconY + ((_iconSize + _iconSpacingY) * y);

			icon->textRect.top = icon->y + _iconSize + _iconTextSpacing;
			icon->textRect.bottom = icon->textRect.top + _iconTextHeight;
			icon->textRect.left = icon->x + (_iconSize / 2) - (_iconTextWidth / 2);
			icon->textRect.right = icon->textRect.left + _iconTextWidth;

			SetRect(&icon->iconRect, icon->x, icon->y, icon->x + _iconSize, icon->y + _iconSize);

			icon->icon = GetIconFromPathEx(icon->filename, _iconSize, _iconSize);
			if(!icon->icon) {
				LSLogPrintf(LOG_ERROR, "IconDesktop", "Could not load icon for \"%s\"", icon->filename);
			}

			i++;
		}

		if(!FindNextFile(search, &fd)) {
			if(GetLastError() == ERROR_NO_MORE_FILES) {
				finished = TRUE;
			}
			else {
				LSLogPrintf(LOG_ERROR, "IconDesktop", "Couldn't find next file.");
			}
		}
	}

	// Close the search handle.
	if(!FindClose(search)) {
		LSLogPrintf(LOG_WARNING, "IconDesktop", "Couldn't close search handle.");
	}

	_iconsLoaded = TRUE;
}

void Desktop::_paintIcons(HDC hDC)
{
	if(_iconsLoaded == FALSE) {
		_loadIcons();
	}

	COLORREF oldColor = SetTextColor(hDC, _iconTextColor);
	HFONT oldFont = (HFONT)SelectObject(hDC, _iconTextFont);
	int oldMode = SetBkMode(hDC, TRANSPARENT);

	DESKTOPICON *loop = _iconList;
	while(loop) {
		if(loop->icon) {
			DrawIconEx(hDC, loop->x, loop->y, loop->icon, _iconSize, _iconSize, 0, NULL, DI_NORMAL);
			if(_iconTextBackground) {
				FillRect(hDC, &loop->textRect, _iconTextBackground);
			}
			if(_selectedIcon == loop) {
				DrawFocusRect(hDC, &loop->textRect);
			}
			DrawText(hDC, loop->title, strlen(loop->title), &loop->textRect, DT_WORDBREAK | DT_CENTER | DT_NOPREFIX /*| DT_NOCLIP*/ | DT_END_ELLIPSIS);
		}
		loop = loop->next;
	}

	SetTextColor(hDC, oldColor);
	SelectObject(hDC, oldFont);
	SetBkMode(hDC, oldMode);
}

DESKTOPICON *Desktop::_getIconAtPt(int x, int y)
{
	POINT mouse = {x, y};
	DESKTOPICON *loop = _iconList;
	while(loop) {
		if(PtInRect(&loop->textRect, mouse) || PtInRect(&loop->iconRect, mouse)) {
			return loop;
		}
		loop = loop->next;
	}
	return NULL;
}


//=========================================================
// Message handlers
//=========================================================

void Desktop::onTimer(Message& message)
{
	if(WaitForSingleObject(_iconChangeHandle, 0) == WAIT_OBJECT_0) {
		_deleteIcons();
		InvalidateRect(hWnd, NULL, TRUE);
		FindNextChangeNotification(_iconChangeHandle);
	} 
}

void Desktop::onPaint(Message& message)
{
	PAINTSTRUCT	ps;
	HDC			hDC;

	if(!message.wParam) {
		hDC = BeginPaint(hWnd, &ps);
	}
	else {
		hDC = (HDC)message.wParam;
	}

	_paintIcons(hDC);

	if(!message.wParam) {
		EndPaint(hWnd, &ps);
	}
}


void Desktop::onActivate(Message& message)
{
	// Tells litestep it has been selected
	SendMessage(hParent, LM_LSSELECT, 0, 0);
	if (message.wParamLo == WA_ACTIVE || message.wParamLo == WA_CLICKACTIVE)
	{
		SetActiveWindow(hParent);
	}
}


void Desktop::onDisplayChange(Message& message)
{
	SetWindowPos(hWnd, HWND_BOTTOM, SCREEN_LEFT, SCREEN_TOP, SCREEN_WIDTH,
	             SCREEN_HEIGHT, SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOSENDCHANGING);
	setMinMax();

	if (bangDisplayChange.length() > 1)
	{
		LSExecute(hWnd, bangDisplayChange.c_str(), 0);
	}
}


void Desktop::onEndSession(Message& message)
{
	message.lResult = SendMessage(hParent, message.uMsg, message.wParam, message.lParam);
}


void Desktop::onEraseBkgnd(Message& message)
{
	PaintDesktop((HDC)(message.wParam));
	message.lResult = TRUE;
}


void Desktop::onGetRevId(Message& message)
{
	LPSTR buf = (LPSTR)(message.lParam);

	switch (message.wParam)
	{
		case 0:
		sprintf(buf, "IconDesktop: %s", &rcsRevision[0]);
		buf[strlen(buf) - 1] = '\0';
		break;
		case 1:
		strcpy(buf, &rcsId[0]);
		buf[strlen(buf) - 1] = '\0';
		break;
		default:
		strcpy(buf, "");
	}
	message.lResult = strlen(buf);
}


void Desktop::onKeyMessage(Message& message)
{
	// Forward these messages
	PostMessage(hParent, message.uMsg, message.wParam, message.lParam);
}


void Desktop::onMouseActivate(Message& message)
{
	// Tells litestep it has been selected
	PostMessage(hParent, LM_LSSELECT, 0, 0);
	// Close any popup open
	//SendMessage(hParent, LM_HIDEPOPUP, (int)(message.lParamHi), (int)(message.lParamLo));
    message.lResult = MA_NOACTIVATE;
}


void Desktop::onMouseButtonDbl(Message& message)
{
	if(message.uMsg == WM_LBUTTONDBLCLK) {
		DESKTOPICON *clicked = _getIconAtPt(message.lParamLo, message.lParamHi);
		if(clicked) {
			LSExecuteEx(NULL, NULL, clicked->filename, NULL, NULL, SW_NORMAL);
		}
	}
	SendMessage(hParent, LM_HIDEPOPUP, (int)(message.lParamHi), (int)(message.lParamLo));
}


void Desktop::onMouseButtonDown(Message& message)
{
	DESKTOPICON *clicked = _getIconAtPt(message.lParamLo, message.lParamHi);
	if(_selectedIcon) {
		InvalidateRect(hWnd, &_selectedIcon->textRect, TRUE);
	}
	_selectedIcon = clicked;
	if(_selectedIcon) {
		InvalidateRect(hWnd, &_selectedIcon->textRect, TRUE);
	}
	SendMessage(hParent, LM_HIDEPOPUP, (int)(message.lParamHi), (int)(message.lParamLo));
}


void Desktop::onMouseButtonUp(Message& message)
{
	DESKTOPICON *clicked = _getIconAtPt(message.lParamLo, message.lParamHi);
	if(clicked) {
		if(message.uMsg == WM_RBUTTONUP) {
			PopupContextMenu(clicked->filename);
		}
		return;
	}

	UINT nModifier = 0;

	switch (message.uMsg)
	{
		case WM_LBUTTONUP:
		nModifier = MOD_LBUTTON;
		break;
		case WM_RBUTTONUP:
		nModifier = MOD_RBUTTON;
		break;
		case WM_MBUTTONUP:
		nModifier = MOD_MBUTTON;
		break;
		case WM_XBUTTONUP:
		switch (GET_XBUTTON_WPARAM(message.wParam))
		{
			case XBUTTON1:
			nModifier = MOD_XBUTTON1;
            break;
			case XBUTTON2:
			nModifier = MOD_XBUTTON2;
			break;
		}
        // in case of the xbuttons the low order word indicates
        // which modifier key was pressed
        message.wParam = GET_KEYSTATE_WPARAM(message.wParam);
        message.lResult = true;
		break;
		default:
		return ;
	}
	nModifier |= message.wParam;

    if (GetKeyState(VK_APPS) < 0)
		nModifier |= MK_APPS;
	if (GetKeyState(VK_MENU) < 0)
		nModifier |= MK_MENU;
	if (GetKeyState(VK_LWIN) < 0)
		nModifier |= MK_LWIN;
	if (GetKeyState(VK_RWIN) < 0)
		nModifier |= MK_RWIN;

	mapClicks::iterator itClicks;
	itClicks = bangClicks.find(nModifier);
	if (itClicks != bangClicks.end())
	{
		LSExecute(hWnd, (*itClicks).second.c_str(), 0);
	}
}


void Desktop::onSysCommand(Message& message)
{
	if (message.wParam == SC_CLOSE)
		PostMessage(hParent, WM_KEYDOWN, LM_SHUTDOWN, 0);
	else
		message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}


void Desktop::onWindowPosChanging(Message& message)
{
	WINDOWPOS *c = (WINDOWPOS*)(message.lParam);

	c->hwndInsertAfter = HWND_BOTTOM;
	c->flags |= SWP_NOMOVE | SWP_NOSIZE;
}

void Desktop::onRefresh(Message& message)
{
	cleanDesktop();

	resetMinMax();

	screenLeft = SCREEN_LEFT;
	screenTop = SCREEN_TOP;
	screenWidth = SCREEN_WIDTH;
	screenHeight = SCREEN_HEIGHT;

	SetWindowPos(hWnd, HWND_BOTTOM, SCREEN_LEFT, SCREEN_TOP, SCREEN_WIDTH,
        SCREEN_HEIGHT, SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOSENDCHANGING);
    
    setupDesktop();
}

//=========================================================
// Bang command handling
//=========================================================

void Desktop::bangSetClick(LPCSTR args)
{
	if (args)
	{
		char	buffer[4096];
		char	token1[4096], extra_text[4096];
		char*	tokens[1];

		tokens[0] = token1;

		strcpy(buffer, args);

		int count;

		token1[0] = extra_text[0] = '\0';

		count = LCTokenize (buffer, tokens, 1, extra_text);

		if (count == 1)
		{

			UINT nModifier = 0;
			nModifier = parseLine(token1);

			mapClicks::iterator itClicks;
			itClicks = bangClicks.find(nModifier);
			if (itClicks != bangClicks.end())
			{
				bangClicks.erase(itClicks);
			}
			bangClicks.insert(mapClicks::value_type(nModifier, extra_text));
		}
	}
}


void Desktop::bangSetArea(LPCSTR args)
{
	char token[MAX_LINE_LENGTH];
	LPCSTR nextToken = args;

	if (GetToken(nextToken, token, &nextToken, FALSE))
		sdaLeft = atoi(token);
	if (GetToken(nextToken, token, &nextToken, FALSE))
		sdaTop = atoi(token);
	if (GetToken(nextToken, token, &nextToken, FALSE))
		sdaRight = atoi(token);
	if (GetToken(nextToken, token, &nextToken, FALSE))
		sdaBottom = atoi(token);
	setMinMax();
}


//Get and set the multiple workareas
BOOL CALLBACK MultipleDesktopAreaEnumProc(HMONITOR hMonitor, HDC gdc, LPRECT prc, LPARAM lParam)
{
	char config[32];
	bool isFoundSDA = false;
	RECT rcSDA;

	// get the exec command
	sprintf( config, "SetDesktopArea%d", numMonitors );
	isFoundSDA = GetRCBool(config, true) != FALSE;
	if (isFoundSDA)
	{
		multipleSDA = true;
		sprintf( config, "SDALeft%d", numMonitors );
		rcSDA.left = GetRCInt(config, 0);
		sprintf( config, "SDARight%d", numMonitors );
		rcSDA.right = GetRCInt(config, 0);
		sprintf( config, "SDATop%d", numMonitors );
		rcSDA.top = GetRCInt(config, 0);
		sprintf( config, "SDABottom%d", numMonitors );
		rcSDA.bottom = GetRCInt(config, 0);

		prc->left += rcSDA.left;
		prc->top += rcSDA.top;
		if (rcSDA.bottom <= 0)
			prc->bottom += rcSDA.bottom;
		else
			prc->bottom = rcSDA.bottom;
		if (rcSDA.right <= 0)
			prc->right += rcSDA.right;
		else
			prc->right = rcSDA.right;

#ifdef _DEBUG

		sprintf(szdebugBuffer, "left: %i top: %i right: %i bottom: %i",
		        prc->left, prc->top, prc->right, prc->bottom);
		_LSDEBUG2("Multiple SDA", szdebugBuffer);
#endif // _DEBUG

		SystemParametersInfo(SPI_SETWORKAREA, 0, (PVOID)prc, SPIF_SENDCHANGE);
		//SystemParametersInfo(SPI_GETWORKAREA, 0, (PVOID)prc, SPIF_SENDCHANGE);
		numMonitors++;
	}

	return true;
}


// Reset the desktop area for each monitor to the max
BOOL CALLBACK DesktopAreaResetEnumProc(HMONITOR hMonitor, HDC gdc, LPRECT prc, LPARAM lParam)
{
#ifdef _DEBUG
	sprintf(szdebugBuffer, "left: %i top: %i right: %i bottom: %i",
	        prc->left, prc->top, prc->right, prc->bottom);
	_LSDEBUG2("resetMinMax", szdebugBuffer);
#endif // _DEBUG

	SystemParametersInfo(SPI_SETWORKAREA, 0, (PVOID)prc, SPIF_SENDCHANGE);
	//SystemParametersInfo(SPI_GETWORKAREA, 0, (PVOID)prc, SPIF_SENDCHANGE);
	return true;
}


// Set the desktop area for everymonitor but the first to the max
BOOL CALLBACK DesktopAreaEnumProc(HMONITOR hMonitor, HDC gdc, LPRECT prc, LPARAM lParam)
{
#ifdef _DEBUG
	sprintf(szdebugBuffer, "left: %i top: %i right: %i bottom: %i",
	        prc->left, prc->top, prc->right, prc->bottom);
	_LSDEBUG2("setMinMax", szdebugBuffer);
#endif // _DEBUG

	PRECT pr = (PRECT)lParam;
	if (!EqualRect(pr, prc))
	{
		SystemParametersInfo(SPI_SETWORKAREA, 0, (PVOID)prc, SPIF_SENDCHANGE);
	}
	//SystemParametersInfo(SPI_GETWORKAREA, 0, (PVOID)prc, SPIF_SENDCHANGE);
	return true;
}
